package fanin_fanout_test

import (
	"context"
	"reflect"
	"sync"
	"testing"
	"time"
)

func TestFanoutOneDestination(t *testing.T) {
	const numMessages = 10

	sender := func(ctx context.Context, ch chan int) {
		defer close(ch)
		for i := 0; i < numMessages; i++ {
			select {
			case <-ctx.Done():
				return
			case ch <- i:
				t.Logf("sent %d", i)
			}
		}
	}
	receiver := func(ctx context.Context, ch chan int, done chan struct{}) {
		defer close(done)
		for i := 0; i < numMessages; i++ {
			select {
			case <-ctx.Done():
				return
			case v := <-ch:
				t.Logf("received %d", v)
			}
		}
	}

	fanoutOneDestination := func(ctx context.Context, src, dst chan int) {
		for {
			data, ok := <-src
			if !ok {
				return
			}
			select {
			case <-ctx.Done():
				return
			case dst <- data:
				t.Logf("relayed %d", data)
			}
		}
	}

	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	src := make(chan int)
	dst := make(chan int)
	done := make(chan struct{})

	go fanoutOneDestination(ctx, src, dst)
	go receiver(ctx, dst, done)
	go sender(ctx, src)

	select {
	case <-ctx.Done():
		t.Errorf("timeout out")
	case <-done:
	}
}

func TestFanoutMultipleStaticDestinations(t *testing.T) {
	const numMessages = 10
	const numDestinations = 3

	sender := func(ctx context.Context, ch chan int) {
		defer close(ch)
		for i := 0; i < numMessages; i++ {
			select {
			case <-ctx.Done():
				return
			case ch <- i:
				t.Logf("sent %d", i)
			}
		}
	}
	receiver := func(ctx context.Context, id int, ch chan int, wg *sync.WaitGroup) {
		defer wg.Done()
		for i := 0; i < numMessages; i++ {
			select {
			case <-ctx.Done():
				return
			case v := <-ch:
				t.Logf("receiver %d received %d", id, v)
			}
		}
	}

	fanoutMultipleStaticDestinations := func(ctx context.Context, src, dst1, dst2, dst3 chan int) {
		for {
			data, ok := <-src
			if !ok {
				return
			}
			select {
			case <-ctx.Done():
				return
			case dst1 <- data:
				t.Logf("relayed %d to dst1", data)
			case dst2 <- data:
				t.Logf("relayed %d to dst2", data)
			case dst3 <- data:
				t.Logf("relayed %d to dst3", data)
			}
		}
	}

	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	src := make(chan int)
	dsts := make([]chan int, 3)
	for i := 0; i < numDestinations; i++ {
		dsts[i] = make(chan int)
	}

	go fanoutMultipleStaticDestinations(ctx, src, dsts[0], dsts[1], dsts[2])
	go sender(ctx, src)
	var wg sync.WaitGroup
	for i := 0; i < numDestinations; i++ {
		wg.Add(1)
		go receiver(ctx, i+1, dsts[i], &wg)
	}

	done := make(chan struct{})
	go func() {
		defer close(done)
		wg.Wait()
	}()

	select {
	case <-ctx.Done():
		t.Errorf("timeout out")
	case <-done:
	}
}

func TestFanoutMultipleStaticDestinationsMultipleWrite(t *testing.T) {
	const numMessages = 10
	const numDestinations = 3

	sender := func(ctx context.Context, ch chan int) {
		defer close(ch)
		for i := 0; i < numMessages; i++ {
			select {
			case <-ctx.Done():
				return
			case ch <- i:
				t.Logf("sent %d", i)
			}
		}
	}
	receiver := func(ctx context.Context, id int, ch chan int, wg *sync.WaitGroup) {
		defer wg.Done()
		for i := 0; i < numMessages; i++ {
			select {
			case <-ctx.Done():
				return
			case v := <-ch:
				t.Logf("receiver %d received %d", id, v)
			}
		}
	}

	fanoutMultipleStaticDestinations := func(ctx context.Context, src, dst1, dst2, dst3 chan int) {
		for {
			data, ok := <-src
			if !ok {
				return
			}
			for i := 0; i < numDestinations; i++ {
				select {
				case <-ctx.Done():
					return
				case dst1 <- data:
					t.Logf("relayed %d to dst1", data)
				case dst2 <- data:
					t.Logf("relayed %d to dst2", data)
				case dst3 <- data:
					t.Logf("relayed %d to dst3", data)
				}
			}
		}
	}

	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	src := make(chan int)
	dsts := make([]chan int, 3)
	for i := 0; i < numDestinations; i++ {
		dsts[i] = make(chan int)
	}

	go fanoutMultipleStaticDestinations(ctx, src, dsts[0], dsts[1], dsts[2])
	go sender(ctx, src)
	var wg sync.WaitGroup
	for i := 0; i < numDestinations; i++ {
		wg.Add(1)
		go receiver(ctx, i+1, dsts[i], &wg)
	}

	done := make(chan struct{})
	go func() {
		defer close(done)
		wg.Wait()
	}()

	select {
	case <-ctx.Done():
		t.Errorf("timeout out")
	case <-done:
	}
}

func TestRotatingDestinations(t *testing.T) {
	const numMessages = 10
	const numDestinations = 3

	sender := func(ctx context.Context, ch chan int) {
		defer close(ch)
		for i := 0; i < numMessages; i++ {
			select {
			case <-ctx.Done():
				return
			case ch <- i:
				t.Logf("sent %d", i)
			}
		}
	}

	receiver := func(ctx context.Context, id int, ch chan int, wg *sync.WaitGroup) {
		defer wg.Done()
		for i := 0; i < numMessages; i++ {
			select {
			case <-ctx.Done():
				return
			case v := <-ch:
				t.Logf("receiver %d received %d", id, v)
			}
		}
	}

	fanoutRotatingDestinations := func(ctx context.Context, src chan int, dsts []chan int) {
		for {
			data, ok := <-src
			if !ok {
				return
			}

			for i, dst := range dsts {
				select {
				case <-ctx.Done():
					return
				case dst <- data:
					t.Logf("relayed %d to dst%d", data, i+1)
				}
			}
		}
	}

	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	src := make(chan int)
	dsts := make([]chan int, 3)
	for i := 0; i < numDestinations; i++ {
		dsts[i] = make(chan int)
	}

	go fanoutRotatingDestinations(ctx, src, dsts)
	go sender(ctx, src)
	var wg sync.WaitGroup
	for i := 0; i < numDestinations; i++ {
		wg.Add(1)
		go receiver(ctx, i+1, dsts[i], &wg)
	}

	done := make(chan struct{})
	go func() {
		defer close(done)
		wg.Wait()
	}()

	select {
	case <-ctx.Done():
		t.Errorf("timeout out")
	case <-done:
	}
}

func TestDynamicDestinations(t *testing.T) {
	const numMessages = 10
	const numDestinations = 3

	sender := func(ctx context.Context, ch chan int) {
		defer close(ch)
		for i := 0; i < numMessages; i++ {
			select {
			case <-ctx.Done():
				return
			case ch <- i:
				t.Logf("sent %d", i)
			}
		}
	}

	receiver := func(ctx context.Context, id int, ch chan int, wg *sync.WaitGroup) {
		defer wg.Done()
		for i := 0; i < numMessages; i++ {
			select {
			case <-ctx.Done():
				return
			case v := <-ch:
				t.Logf("receiver %d received %d", id, v)
			}
		}
	}

	fanoutRotatingDestinations := func(ctx context.Context, src chan int, dsts []chan int) {
		for {
			data, ok := <-src
			if !ok {
				return
			}

			cases := make([]reflect.SelectCase, len(dsts)+1)
			for i, dst := range dsts {
				cases[i] = reflect.SelectCase{
					Chan: reflect.ValueOf(dst),
					Dir:  reflect.SelectSend,
					Send: reflect.ValueOf(data),
				}
			}
			cases[len(dsts)] = reflect.SelectCase{
				Chan: reflect.ValueOf(ctx.Done()),
				Dir:  reflect.SelectRecv,
			}

			for len(cases) > 1 {
				chosen, _, _ := reflect.Select(cases)
				if chosen == len(cases)-1 { // ctx.Done()
					return
				}

				cases = append(cases[:chosen], cases[chosen+1:]...)
			}
		}
	}

	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	src := make(chan int)
	dsts := make([]chan int, 3)
	for i := 0; i < numDestinations; i++ {
		dsts[i] = make(chan int)
	}

	go fanoutRotatingDestinations(ctx, src, dsts)
	go sender(ctx, src)
	var wg sync.WaitGroup
	for i := 0; i < numDestinations; i++ {
		wg.Add(1)
		go receiver(ctx, i+1, dsts[i], &wg)
	}

	done := make(chan struct{})
	go func() {
		defer close(done)
		wg.Wait()
	}()

	select {
	case <-ctx.Done():
		t.Errorf("timeout out")
	case <-done:
	}
}

func TestFaninDynamic(t *testing.T) {
	const numMessages = 10
	const numSenders = 3
	sender := func(ctx context.Context, id int, ch chan int, wg *sync.WaitGroup) {
		defer wg.Done()
		defer close(ch)

		for i := 0; i < 10; i++ {
			select {
			case <-ctx.Done():
				return
			case ch <- i:
				t.Logf("sender %d sent %d", id, i)
			}
		}
	}
	receiver := func(ctx context.Context, ch chan int) {
		for i := 0; i < numMessages*numSenders; i++ {
			select {
			case <-ctx.Done():
				return
			case v, ok := <-ch:
				if !ok {
					continue
				}
				t.Logf("received %d", v)
			}
		}
	}
	faninDynamic := func(ctx context.Context, srcs []chan int, dst chan int) {
		cases := make([]reflect.SelectCase, len(srcs)+1)
		for i, src := range srcs {
			cases[i] = reflect.SelectCase{
				Dir:  reflect.SelectRecv,
				Chan: reflect.ValueOf(src),
			}
		}
		cases[len(srcs)] = reflect.SelectCase{
			Dir:  reflect.SelectRecv,
			Chan: reflect.ValueOf(ctx.Done()),
		}

		for {
			chosen, recv, recvOK := reflect.Select(cases)
			if chosen == len(srcs) {
				return
			}
			if recvOK {
				dst <- recv.Interface().(int)
			}
		}
	}

	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	srcs := make([]chan int, numSenders)
	for i := 0; i < numSenders; i++ {
		srcs[i] = make(chan int)
	}

	dst := make(chan int)

	go faninDynamic(ctx, srcs, dst)

	var wg sync.WaitGroup
	for i := 0; i < numSenders; i++ {
		wg.Add(1)
		go sender(ctx, i+1, srcs[i], &wg)
	}

	go receiver(ctx, dst)

	done := make(chan struct{})
	go func() {
		defer close(done)
		wg.Wait()
	}()

	select {
	case <-ctx.Done():
		t.Errorf("timeout out")
	case <-done:
	}
}
